use base64::Engine;
use fastly::error::anyhow;
use fastly::http::{header, HeaderValue, Method, StatusCode};
use fastly::{mime, ConfigStore, Error, Request, Response};
use hex::FromHex;
use serde::Deserialize;
use serde_json::json;

type Result<T> = core::result::Result<T, Error>;

struct UpstashRedisBackend {
    url: String,
    key: String,
}

#[derive(Deserialize)]
enum UpstashResponse {
    #[serde(rename = "result")]
    Ok(serde_json::Value),
    #[serde(rename = "error")]
    Err(String),
}

impl From<UpstashResponse> for Result<serde_json::Value> {
    fn from(value: UpstashResponse) -> Self {
        match value {
            UpstashResponse::Ok(value) => Ok(value),
            UpstashResponse::Err(err) => Err(anyhow!("Upstash Redis error: {}", err)),
        }
    }
}

impl StorageBackend for UpstashRedisBackend {
    fn from_configstore(store: &ConfigStore) -> Self {
        Self {
            url: store.get("upstash_redis_url").unwrap(),
            key: store.get("upstash_redis_key").unwrap(),
        }
    }

    fn get_modified(&mut self, id: u64) -> Result<Option<u64>> {
        let id_hex = hex::encode(id.to_le_bytes());
        let response: Result<serde_json::Value> = Request::new(
            Method::GET,
            format!("{}/HGET/{}/last_modified", self.url, id_hex),
        )
        .with_header(header::AUTHORIZATION, format!("Bearer {}", self.key))
        .send("upstash_redis")?
        .take_body_json::<UpstashResponse>()?
        .into();
        let response: Option<String> = serde_json::from_value(response?)?;
        Ok(if let Some(response) = response {
            let timestamp = u64::from_le_bytes(<[u8; 8]>::from_hex(response)?);
            Some(timestamp)
        } else {
            None
        })
    }

    fn get(&mut self, id: u64) -> Result<Option<(Vec<u8>, u64)>> {
        let id_hex = hex::encode(id.to_le_bytes());
        let response: Result<serde_json::Value> = Request::new(
            Method::GET,
            format!("{}/HMGET/{}/last_modified/settings", self.url, id_hex),
        )
        .with_header(header::AUTHORIZATION, format!("Bearer {}", self.key))
        .send("upstash_redis")?
        .take_body_json::<UpstashResponse>()?
        .into();
        let response: (Option<String>, Option<String>) = serde_json::from_value(response?)?;
        Ok(if let (Some(timestamp), Some(data)) = response {
            let timestamp = u64::from_le_bytes(<[u8; 8]>::from_hex(timestamp)?);
            let data = BASE64_ENGINE.decode(data)?;
            Some((data, timestamp))
        } else {
            None
        })
    }

    fn set(&mut self, id: u64, data: Vec<u8>) -> Result<u64> {
        let now = time::OffsetDateTime::now_utc();
        let now = (now.unix_timestamp_nanos() / 1000000) as u64;

        let id_hex = hex::encode(id.to_le_bytes());
        let now_hex = hex::encode(now.to_le_bytes());

        let data_base64 = BASE64_ENGINE.encode(data);

        let response: Result<serde_json::Value> = Request::new(
            Method::GET,
            format!(
                "{}/HSET/{}/last_modified/{}/settings",
                self.url, id_hex, now_hex
            ),
        )
        .with_header(header::AUTHORIZATION, format!("Bearer {}", self.key))
        .with_body_octet_stream(data_base64.as_bytes())
        .send("upstash_redis")?
        .take_body_json::<UpstashResponse>()?
        .into();
        response.map(|_| now)
    }

    fn del(&mut self, id: u64) -> Result<()> {
        let id_hex = hex::encode(id.to_le_bytes());
        let response: Result<serde_json::Value> = Request::new(
            Method::GET,
            format!("{}/HDEL/{}/last_modified/settings", self.url, id_hex),
        )
        .with_header(header::AUTHORIZATION, format!("Bearer {}", self.key))
        .send("upstash_redis")?
        .take_body_json::<UpstashResponse>()?
        .into();
        response.map(|_| ())
    }
}

type DefaultStorageBackend = UpstashRedisBackend;

trait StorageBackend {
    fn from_configstore(store: &ConfigStore) -> Self;
    fn get_modified(&mut self, id: u64) -> Result<Option<u64>>;
    fn get(&mut self, id: u64) -> Result<Option<(Vec<u8>, u64)>>;
    fn set(&mut self, id: u64, data: Vec<u8>) -> Result<u64>;
    fn del(&mut self, id: u64) -> Result<()>;
}

const ALLOWED_DISCORD_ORIGINS: [&[u8]; 3] = [
    b"https://discord.com",
    b"https://ptb.discord.com",
    b"https://canary.discord.com",
];

const DISCORD_OAUTH_TOKEN_URL: &str = "https://discord.com/api/oauth2/token";
const DISCORD_SELF_INFO_URL: &str = "https://discord.com/api/users/@me";

const BASE64_ENGINE: base64::engine::GeneralPurpose =
    base64::engine::general_purpose::STANDARD_NO_PAD;

fn handle_with_cors(
    request: Request,
    methods: &'static str,
    allowed_headers: &'static str,
    mut handler: impl FnMut(Request) -> Result<Response>,
) -> Result<Response> {
    Ok(if let &Method::OPTIONS = request.get_method() {
        match request.get_header(header::ORIGIN) {
            Some(origin) => {
                if ALLOWED_DISCORD_ORIGINS.contains(&origin.as_bytes()) {
                    Response::from_status(StatusCode::OK)
                        .with_header(header::ACCESS_CONTROL_ALLOW_ORIGIN, origin)
                        .with_header(header::ACCESS_CONTROL_ALLOW_HEADERS, allowed_headers)
                        .with_header(header::ACCESS_CONTROL_ALLOW_METHODS, methods)
                        .with_header(header::VARY, "Origin")
                } else {
                    Response::from_status(StatusCode::FORBIDDEN)
                }
            }
            None => Response::from_status(StatusCode::OK).with_header(header::ALLOW, methods),
        }
    } else {
        let origin = request.get_header(header::ORIGIN).cloned();
        let mut response = handler(request)?;
        if response.get_status() == StatusCode::METHOD_NOT_ALLOWED {
            response.set_header(header::ALLOW, methods);
        } else if let Some(origin) = origin {
            response.set_header(header::ACCESS_CONTROL_ALLOW_ORIGIN, origin);
            response.set_header(header::ACCESS_CONTROL_ALLOW_HEADERS, allowed_headers);
            response.set_header(header::ACCESS_CONTROL_ALLOW_METHODS, methods);
            response.set_header(header::VARY, "Origin");
        }
        response
    })
}

#[derive(Deserialize)]
struct TokenResponse {
    access_token: String,
}

#[derive(Deserialize)]
struct UserInfo {
    id: String,
}

fn check_auth(token: &HeaderValue, secret: &[u8; 32]) -> Option<u64> {
    let decoded = base64::engine::general_purpose::STANDARD.decode(token).ok()?;
    let decoded = String::from_utf8(decoded).ok()?;
    let token = decoded.split_terminator(':').next()?;
    let decoded = BASE64_ENGINE.decode(token).ok()?;
    if decoded.len() != 40 {
        return None;
    }
    let (id, mac) = decoded.split_at(8);
    let id = u64::from_le_bytes(id.try_into().unwrap());
    let mac: [u8; 32] = mac.try_into().unwrap();
    let mac: blake3::Hash = mac.into();
    let correct_mac = blake3::keyed_hash(secret, &id.to_le_bytes());
    if mac == correct_mac {
        Some(id)
    } else {
        println!("mac is wrong; got {:?}, expected {:?}", mac, correct_mac);
        None
    }
}

#[fastly::main]
fn main(req: Request) -> Result<Response> {
    let config = ConfigStore::open("synkronisera");

    let client_id = config.get("client_id").unwrap();
    let client_secret = config.get("client_secret").unwrap();
    let redirect_url = config.get("redirect_url").unwrap();

    let mac_secret = <[u8; 32]>::from_hex(config.get("mac_secret").unwrap()).unwrap();
    let max_config_size: usize = config.get("max_config_size").unwrap().parse().unwrap();

    let mut store = DefaultStorageBackend::from_configstore(&config);

    match req.get_path() {
        "/" => Ok(Response::from_status(StatusCode::OK)
            .with_content_type(mime::TEXT_HTML_UTF_8)
            .with_body(include_str!("index.html"))),

        "/v1" => handle_with_cors(req, "OPTIONS, GET, HEAD", "Authorization", |req| {
            Ok(match *req.get_method() {
                Method::GET => Response::from_status(StatusCode::OK).with_body_json(&json!({
                    "ping": "pong"
                }))?,
                Method::HEAD => Response::from_status(StatusCode::OK),
                _ => Response::from_status(StatusCode::METHOD_NOT_ALLOWED),
            })
        }),

        "/v1/" => handle_with_cors(req, "OPTIONS, DELETE", "Authorization", |req| {
            Ok(match *req.get_method() {
                Method::DELETE => {
                    let id = if let Some(id) = req
                        .get_header(header::AUTHORIZATION)
                        .and_then(|token| check_auth(token, &mac_secret))
                    {
                        id
                    } else {
                        return Ok(Response::from_status(StatusCode::UNAUTHORIZED)
                            .with_body_json(&json!({
                                "error": "Unauthorized"
                            }))?);
                    };
                    store.del(id)?;
                    Response::from_status(StatusCode::NO_CONTENT)
                }
                _ => Response::from_status(StatusCode::METHOD_NOT_ALLOWED),
            })
        }),

        "/v1/oauth/settings" => handle_with_cors(req, "OPTIONS, GET, HEAD", "Authorization", |req| {
            Ok(match *req.get_method() {
                Method::GET => Response::from_status(StatusCode::OK).with_body_json(&json!({
                    "clientId": client_id,
                    "redirectUri": redirect_url
                }))?,
                Method::HEAD => Response::from_status(StatusCode::OK),
                _ => Response::from_status(StatusCode::METHOD_NOT_ALLOWED),
            })
        }),

        "/v1/oauth/callback" => handle_with_cors(req, "OPTIONS, GET", "Authorization", |req| {
            Ok(match *req.get_method() {
                Method::GET => {
                    if let Some(code) = req.get_query_parameter("code") {
                        let mut token_response =
                            Request::new(Method::POST, DISCORD_OAUTH_TOKEN_URL)
                                .with_body_form(&json!({
                                    "client_id": client_id,
                                    "client_secret": client_secret,
                                    "grant_type": "authorization_code",
                                    "code": code,
                                    "redirect_uri": redirect_url,
                                    "scope": "identify"
                                }))?
                                .send("discord")?;
                        if token_response.get_status() != StatusCode::OK {
                            Response::from_status(StatusCode::UNAUTHORIZED).with_body_json(
                                &json!({
                                    "error": "Failed to authorize with Discord"
                                }),
                            )?
                        } else {
                            let token = token_response
                                .take_body_json::<TokenResponse>()?
                                .access_token;
                            let user_id: u64 = Request::new(Method::GET, DISCORD_SELF_INFO_URL)
                                .with_header(header::AUTHORIZATION, format!("Bearer {}", token))
                                .send("discord")?
                                .take_body_json::<UserInfo>()?
                                .id
                                .parse()?;
                            let user_id_bytes = user_id.to_le_bytes();
                            let mac: [u8; 32] =
                                blake3::keyed_hash(&mac_secret, &user_id_bytes).into();
                            let combined_token = [
                                user_id_bytes[0],
                                user_id_bytes[1],
                                user_id_bytes[2],
                                user_id_bytes[3],
                                user_id_bytes[4],
                                user_id_bytes[5],
                                user_id_bytes[6],
                                user_id_bytes[7],
                                mac[0],
                                mac[1],
                                mac[2],
                                mac[3],
                                mac[4],
                                mac[5],
                                mac[6],
                                mac[7],
                                mac[8],
                                mac[9],
                                mac[10],
                                mac[11],
                                mac[12],
                                mac[13],
                                mac[14],
                                mac[15],
                                mac[16],
                                mac[17],
                                mac[18],
                                mac[19],
                                mac[20],
                                mac[21],
                                mac[22],
                                mac[23],
                                mac[24],
                                mac[25],
                                mac[26],
                                mac[27],
                                mac[28],
                                mac[29],
                                mac[30],
                                mac[31],
                            ];
                            Response::from_status(StatusCode::OK).with_body_json(&json!({
                                "secret": BASE64_ENGINE.encode(combined_token)
                            }))?
                        }
                    } else {
                        Response::from_status(StatusCode::BAD_REQUEST).with_body_json(&json!({
                            "error": "Missing code"
                        }))?
                    }
                }
                _ => Response::from_status(StatusCode::METHOD_NOT_ALLOWED),
            })
        }),

        "/v1/settings" => handle_with_cors(req, "OPTIONS, GET, HEAD, PUT, DELETE", "Authorization, Content-Type, If-None-Match", |req| {
            let id = if let Some(id) = req
                .get_header(header::AUTHORIZATION)
                .and_then(|token| check_auth(token, &mac_secret))
            {
                id
            } else {
                return Ok(
                    Response::from_status(StatusCode::UNAUTHORIZED).with_body_json(&json!({
                        "error": "Unauthorized"
                    }))?,
                );
            };
            Ok(match *req.get_method() {
                Method::GET => {
                    if let Some((data, timestamp)) = store.get(id)? {
                        fn get_cached_timestamp(header: Option<&HeaderValue>) -> Option<u64> {
                            header?.to_str().ok()?.parse().ok()
                        }

                        if let Some(cached_timestamp) =
                            get_cached_timestamp(req.get_header(header::IF_NONE_MATCH))
                        {
                            if cached_timestamp == timestamp {
                                return Ok(Response::from_status(StatusCode::NOT_MODIFIED)
                                    .with_header(header::ETAG, format!("{}", timestamp)));
                            }
                        }

                        Response::from_status(StatusCode::OK)
                            .with_header(header::ETAG, format!("{}", timestamp))
                            .with_body_octet_stream(&data)
                    } else {
                        Response::from_status(StatusCode::NOT_FOUND)
                    }
                }
                Method::HEAD => {
                    if let Some(timestamp) = store.get_modified(id)? {
                        fn get_cached_timestamp(header: Option<&HeaderValue>) -> Option<u64> {
                            header?.to_str().ok()?.parse().ok()
                        }

                        if let Some(cached_timestamp) =
                            get_cached_timestamp(req.get_header(header::IF_NONE_MATCH))
                        {
                            if cached_timestamp == timestamp {
                                return Ok(Response::from_status(StatusCode::NOT_MODIFIED)
                                    .with_header(header::ETAG, format!("{}", timestamp)));
                            }
                        }

                        Response::from_status(StatusCode::OK)
                            .with_header(header::ETAG, format!("{}", timestamp))
                            .with_header(header::CONTENT_TYPE, "application/octet-stream")
                    } else {
                        Response::from_status(StatusCode::NOT_FOUND)
                    }
                }
                Method::PUT => {
                    if Some(b"application/octet-stream" as &[u8])
                        != req
                            .get_header(header::CONTENT_TYPE)
                            .map(|header| header.as_bytes())
                    {
                        return Ok(Response::from_status(StatusCode::UNSUPPORTED_MEDIA_TYPE));
                    }
                    let body = req.into_body_bytes();
                    if body.len() > max_config_size {
                        return Ok(Response::from_status(StatusCode::PAYLOAD_TOO_LARGE));
                    }
                    let etag = store.set(id, body)?;
                    Response::from_status(StatusCode::OK).with_body_json(&json!({
                        "written": etag
                    }))?
                }
                Method::DELETE => {
                    store.del(id)?;
                    Response::from_status(StatusCode::NO_CONTENT)
                }
                _ => Response::from_status(StatusCode::METHOD_NOT_ALLOWED),
            })
        }),

        _ => Ok(Response::from_status(StatusCode::NOT_FOUND)
            .with_body_text_plain("The page you requested could not be found\n")),
    }
}
